﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  /// <summary>
  /// 
  /// </summary>
  [StorableClass]
  [Item("NeighbourhoodWeightCalculator", "")]
  public class NeighbourhoodWeightCalculator : DiscriminantClassificationWeightCalculator {

    public NeighbourhoodWeightCalculator()
      : base() {
    }

    [StorableConstructor]
    protected NeighbourhoodWeightCalculator(bool deserializing) : base(deserializing) { }
    protected NeighbourhoodWeightCalculator(NeighbourhoodWeightCalculator original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new NeighbourhoodWeightCalculator(this, cloner);
    }

    protected override IEnumerable<double> DiscriminantCalculateWeights(IEnumerable<IDiscriminantFunctionClassificationSolution> discriminantSolutions) {
      List<List<double>> estimatedValues = new List<List<double>>();
      List<List<double>> estimatedClassValues = new List<List<double>>();

      List<IClassificationProblemData> solutionProblemData = discriminantSolutions.Select(sol => sol.ProblemData).ToList();
      Dataset dataSet = solutionProblemData[0].Dataset;
      IEnumerable<int> rows = Enumerable.Range(0, dataSet.Rows);
      foreach (var solution in discriminantSolutions) {
        estimatedValues.Add(solution.Model.GetEstimatedValues(dataSet, rows).ToList());
        estimatedClassValues.Add(solution.Model.GetEstimatedClassValues(dataSet, rows).ToList());
      }

      List<double> weights = Enumerable.Repeat<double>(0, solutionProblemData.Count).ToList<double>();
      List<double> targetValues = dataSet.GetDoubleValues(solutionProblemData[0].TargetVariable).ToList();

      double pointAvg, help;
      int count;
      for (int point = 0; point < targetValues.Count; point++) {
        pointAvg = 0.0;
        count = 0;
        for (int solutionPos = 0; solutionPos < estimatedClassValues.Count; solutionPos++) {
          if (PointInTraining(solutionProblemData[solutionPos], point)
              && estimatedClassValues[solutionPos][point].Equals(targetValues[point])) {
            pointAvg += estimatedValues[solutionPos][point];
            count++;
          }
        }
        pointAvg /= (double)count;
        for (int solutionPos = 0; solutionPos < estimatedClassValues.Count; solutionPos++) {
          if (PointInTraining(solutionProblemData[solutionPos], point)
              && estimatedClassValues[solutionPos][point].Equals(targetValues[point])) {
            weights[solutionPos] += 0.5;
            help = Math.Abs(estimatedValues[solutionPos][point] - 0.5);
            weights[solutionPos] += help < 0.5 ? 0.5 - help : 0.0;
          }
        }
      }
      // normalize the weight (otherwise a model with a bigger training partition would probably be better)
      for (int i = 0; i < weights.Count; i++) {
        weights[i] = weights[i] / solutionProblemData[i].TrainingIndizes.Count();
      }
      return weights;
    }
  }
}
